package envoy.client.ui;

import java.awt.Color;
import java.awt.ComponentOrientation;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.logging.Logger;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.border.EmptyBorder;

import envoy.client.Chat;
import envoy.client.Client;
import envoy.client.Config;
import envoy.client.LocalDB;
import envoy.schema.Message;
import envoy.schema.Sync;
import envoy.schema.User;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>ChatWindow.java</strong><br>
 * Created: <strong>28 Sep 2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @author Leon Hofmeister
 * @since Envoy v0.1-alpha
 */
public class ChatWindow extends JFrame {

	private static final long serialVersionUID = 6865098428255463649L;

	private JPanel contentPane = new JPanel();

	private Client	client;
	private LocalDB	localDB;

	private JList<User>	userList	= new JList<>();
	private Chat		currentChat;

	private JTextArea messageEnterTextArea;
	
	private static final Logger logger = Logger.getLogger(ChatWindow.class.getSimpleName());

	public ChatWindow(Client client, LocalDB localDB) {
		this.client		= client;
		this.localDB	= localDB;

		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setBounds(100, 100, 600, 800);
		setTitle("Envoy");
		setLocationRelativeTo(null);

		// Save chats when window closes
		addWindowListener(new WindowAdapter() {

			@Override
			public void windowClosing(WindowEvent e) { localDB.saveToLocalDB(); }
		});

		contentPane.setBackground(new Color(0, 0, 0));
		contentPane.setForeground(Color.white);
		contentPane.setBorder(new EmptyBorder(0, 5, 0, 0));
		setContentPane(contentPane);
		GridBagLayout gbl_contentPane = new GridBagLayout();
		gbl_contentPane.columnWidths	= new int[] { 1, 1, 1 };
		gbl_contentPane.rowHeights		= new int[] { 1, 1, 1 };
		gbl_contentPane.columnWeights	= new double[] { 0.3, 1.0, 0.1 };
		gbl_contentPane.rowWeights		= new double[] { 0.05, 1.0, 0.07 };
		contentPane.setLayout(gbl_contentPane);

		JList<Message> messageList = new JList<>();
		messageList.setCellRenderer(new MessageListRenderer());

		messageList.setFocusTraversalKeysEnabled(false);
		messageList.setSelectionForeground(new Color(255, 255, 255));
		messageList.setSelectionBackground(new Color(102, 0, 153));
		messageList.setComponentOrientation(ComponentOrientation.LEFT_TO_RIGHT);
		messageList.setForeground(new Color(255, 255, 255));
		messageList.setBackground(new Color(51, 51, 51));

		DefaultListModel<Message> messageListModel = new DefaultListModel<>();
		messageList.setModel(messageListModel);
		messageList.setFont(new Font("Arial", Font.PLAIN, 17));
		messageList.setFixedCellHeight(60);
		messageList.setBorder(new EmptyBorder(5, 5, 5, 5));

		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setForeground(new Color(0, 0, 0));
		scrollPane.setBackground(new Color(51, 51, 51));
		scrollPane.setViewportView(messageList);
		scrollPane.setBorder(null);

		GridBagConstraints gbc_scrollPane = new GridBagConstraints();
		gbc_scrollPane.fill			= GridBagConstraints.BOTH;
		gbc_scrollPane.gridwidth	= 2;
		gbc_scrollPane.gridx		= 1;
		gbc_scrollPane.gridy		= 1;

		gbc_scrollPane.insets = new Insets(0, 10, 10, 10);

		contentPane.add(scrollPane, gbc_scrollPane);

		// Message enter field
		messageEnterTextArea = new JTextArea();
		messageEnterTextArea.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if (e.getKeyCode() == KeyEvent.VK_ENTER
						&& ((SettingsScreen.enterToSend && e.getModifiersEx() == 0) || (e.getModifiersEx() == KeyEvent.CTRL_DOWN_MASK))) {

					postMessage(messageList);
				}
			}
		});
		// Checks for changed Message
		messageEnterTextArea.setWrapStyleWord(true);
		messageEnterTextArea.setCaretColor(new Color(255, 255, 255));
		messageEnterTextArea.setForeground(new Color(255, 255, 255));
		messageEnterTextArea.setBackground(new Color(51, 51, 51));
		messageEnterTextArea.setLineWrap(true);
		messageEnterTextArea.setBorder(null);
		messageEnterTextArea.setFont(new Font("Arial", Font.PLAIN, 17));
		messageEnterTextArea.setBorder(new EmptyBorder(5, 5, 5, 5));

		GridBagConstraints gbc_messageEnterTextfield = new GridBagConstraints();
		gbc_messageEnterTextfield.fill	= GridBagConstraints.BOTH;
		gbc_messageEnterTextfield.gridx	= 1;
		gbc_messageEnterTextfield.gridy	= 2;

		gbc_messageEnterTextfield.insets = new Insets(10, 10, 10, 10);

		contentPane.add(messageEnterTextArea, gbc_messageEnterTextfield);

		// Post Button
		JButton postButton = new JButton("Post");
		postButton.setForeground(new Color(255, 255, 255));
		postButton.setBackground(new Color(102, 51, 153));
		postButton.setBorderPainted(false);

		GridBagConstraints gbc_moveSelectionPostButton = new GridBagConstraints();

		gbc_moveSelectionPostButton.fill	= GridBagConstraints.BOTH;
		gbc_moveSelectionPostButton.gridx	= 2;
		gbc_moveSelectionPostButton.gridy	= 2;

		gbc_moveSelectionPostButton.insets = new Insets(10, 10, 10, 10);

		postButton.addActionListener((evt) -> { postMessage(messageList); });
		contentPane.add(postButton, gbc_moveSelectionPostButton);

		// Settings Button
		JButton settingsButton = new JButton("Settings");
		settingsButton.setForeground(new Color(255, 255, 255));
		settingsButton.setBackground(new Color(102, 51, 153));
		settingsButton.setBorderPainted(false);

		GridBagConstraints gbc_moveSelectionSettingsButton = new GridBagConstraints();

		gbc_moveSelectionSettingsButton.fill	= GridBagConstraints.BOTH;
		gbc_moveSelectionSettingsButton.gridx	= 2;
		gbc_moveSelectionSettingsButton.gridy	= 0;

		gbc_moveSelectionSettingsButton.insets = new Insets(10, 10, 10, 10);

		settingsButton.addActionListener((evt) -> {
			try {
				SettingsScreen.open(localDB.getUser().getName());
			} catch (Exception e) {
				SettingsScreen.open();
				logger.warning("An error occured while opening the settings screen: " + e);
				e.printStackTrace();
			}
		});
		contentPane.add(settingsButton, gbc_moveSelectionSettingsButton);

		// Partner name display
		JTextPane textPane = new JTextPane();
		textPane.setBackground(new Color(0, 0, 0));
		textPane.setForeground(new Color(255, 255, 255));

		textPane.setFont(new Font("Arial", Font.PLAIN, 20));

		GridBagConstraints gbc_partnerName = new GridBagConstraints();
		gbc_partnerName.fill	= GridBagConstraints.HORIZONTAL;
		gbc_partnerName.gridx	= 1;
		gbc_partnerName.gridy	= 0;

		gbc_partnerName.insets = new Insets(0, 10, 0, 10);
		contentPane.add(textPane, gbc_partnerName);

		userList.setCellRenderer(new UserListRenderer());
		userList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		userList.addListSelectionListener((listSelectionEvent) -> {
			if (!listSelectionEvent.getValueIsAdjusting()) {
				@SuppressWarnings("unchecked")
				final JList<User>	selectedUserList	= (JList<User>) listSelectionEvent.getSource();
				final User			user				= selectedUserList.getSelectedValue();
				client.setRecipient(user);

				currentChat = localDB.getChats().stream().filter(chat -> chat.getRecipient().getID() == user.getID()).findFirst().get();

				// Set all unread messages in the chat to read
				readCurrentChat();

				client.setRecipient(user);

				textPane.setText(currentChat.getRecipient().getName());

				messageList.setModel(currentChat.getModel());
				contentPane.revalidate();
			}
		});

		userList.setSelectionForeground(new Color(255, 255, 255));
		userList.setSelectionBackground(new Color(102, 0, 153));
		userList.setForeground(new Color(255, 255, 255));
		userList.setBackground(new Color(51, 51, 51));
		userList.setFont(new Font("Arial", Font.PLAIN, 17));
		userList.setBorder(new EmptyBorder(5, 5, 5, 5));

		GridBagConstraints gbc_userList = new GridBagConstraints();
		gbc_userList.fill	= GridBagConstraints.VERTICAL;
		gbc_userList.gridx	= 0;
		gbc_userList.gridy	= 1;
		gbc_userList.anchor	= GridBagConstraints.PAGE_START;
		gbc_userList.insets	= new Insets(0, 0, 10, 0);

		contentPane.add(userList, gbc_userList);
		contentPane.revalidate();

		loadUsersAndChats();
		startSyncThread(Config.getInstance().getSyncTimeout());

		contentPane.revalidate();
	}

	private void postMessage(JList<Message> messageList) {
		if (!client.hasRecipient()) {
			JOptionPane.showMessageDialog(this, "Please select a recipient!", "Cannot send message", JOptionPane.INFORMATION_MESSAGE);
		}

		if (!messageEnterTextArea.getText().isEmpty()) try {

			// Create and send message object
			final Message message = localDB.createMessage(messageEnterTextArea.getText(), currentChat.getRecipient());
			currentChat.appendMessage(message);
			messageList.setModel(currentChat.getModel());

			// Clear text field
			messageEnterTextArea.setText("");
			contentPane.revalidate();
		} catch (Exception e) {
			JOptionPane.showMessageDialog(this,
					"An exception occured while sending a message. See the log for more details.",
					"Exception occured",
					JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		}
	}

	/**
	 * Initializes the elements of the user list by downloading them from the
	 * server.
	 *
	 * @since Envoy v0.1-alpha
	 */
	private void loadUsersAndChats() {
		new Thread(() -> {
			Sync					users			= client.getUsersListXml();
			DefaultListModel<User>	userListModel	= new DefaultListModel<>();
			users.getUsers().forEach(user -> {
				userListModel.addElement(user);

				// Check if user exists in local DB
				if (localDB.getChats().stream().filter(c -> c.getRecipient().getID() == user.getID()).count() == 0)
					localDB.getChats().add(new Chat(user));
			});
			SwingUtilities.invokeLater(() -> userList.setModel(userListModel));
		}).start();
	}

	/**
	 * Updates the data model and the UI repeatedly after a certain amount of
	 * time.
	 *
	 * @param timeout the amount of time that passes between two requests sent to
	 *                the server
	 * @since Envoy v0.1-alpha
	 */
	private void startSyncThread(int timeout) {
		new Timer(timeout, (evt) -> {
			new Thread(() -> {

				// Synchronize
				localDB.applySync(client.sendSync(client.getSender().getID(), localDB.fillSync(client.getSender().getID())));

				// Process unread messages
				localDB.addUnreadMessagesToLocalDB();
				localDB.clearUnreadMessagesSync();

				// Mark unread messages as read when they are in the current chat
				readCurrentChat();

				// Update UI
				SwingUtilities.invokeLater(() -> { updateUserStates(); contentPane.revalidate(); contentPane.repaint(); });
			}).start();
		}).start();
	}

	private void updateUserStates() {
		for (int i = 0; i < userList.getModel().getSize(); i++)
			for (int j = 0; j < localDB.getChats().size(); j++)
				if (userList.getModel().getElementAt(i).getID() == localDB.getChats().get(j).getRecipient().getID())
					userList.getModel().getElementAt(i).setStatus(localDB.getChats().get(j).getRecipient().getStatus());
	}

	/**
	 * Marks messages in the current chat as {@code READ}.
	 */
	private void readCurrentChat() { if (currentChat != null) { localDB.setMessagesToRead(currentChat); } }
}
